/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 David Zeuthen <david@fubar.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <glib-object.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <polkit/polkit.h>
#include <polkit-dbus/polkit-dbus.h>

static DBusConnection *system_bus;
static PolKitContext *pk_context;


static DBusGProxy *polkit_gnome_proxy;

static DBusGConnection *
get_session_bus (void)
{
        GError          *error;
        DBusGConnection *bus;
        DBusConnection  *connection;

        error = NULL;
        bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
        if (bus == NULL) {
                g_warning ("Couldn't connect to session bus: %s", error->message);
                g_error_free (error);
                goto out;
        }

        connection = dbus_g_connection_get_connection (bus);
 out:
        return bus;
}

static GtkWidget *toplevel_window;
static GtkWidget *button_frobnicate;
static GtkWidget *button_tweak;
static GtkWidget *button_twiddle;
static GtkWidget *button_punch;
static GtkWidget *kick_combo_box;
static GtkWidget *button_kick;

static void
simple_button_cb (GtkButton *button, gpointer user_data)
{
        int status;
        int return_code;
        char *helper_stderr = NULL;
        GError *error = NULL;
        gboolean gained_privilege;
        char *helper_cmdline;
        char *cmdline;

        /* gross hack - but it works */
        if (button == GTK_BUTTON (button_frobnicate)) {
                helper_cmdline = g_strdup ("polkit-gnome-example-helper --frobnicate");
        } else if (button == GTK_BUTTON (button_tweak)) {
                helper_cmdline = g_strdup ("polkit-gnome-example-helper --tweak");
        } else if (button == GTK_BUTTON (button_twiddle)) {
                helper_cmdline = g_strdup ("polkit-gnome-example-helper --twiddle");
        } else if (button == GTK_BUTTON (button_punch)) {
                helper_cmdline = g_strdup ("polkit-gnome-example-helper --punch");
        } else if (button == GTK_BUTTON (button_kick)) {
                helper_cmdline = g_strdup_printf ("polkit-gnome-example-helper --kick %s",
                                                  gtk_combo_box_get_active_text (GTK_COMBO_BOX (kick_combo_box)));
        } else {
                g_warning ("Unknown button pressed");
                goto out;
        }

        cmdline = g_strdup_printf (LIBEXECDIR /* "." */ "/%s", helper_cmdline);
        g_free (helper_cmdline);
        g_debug ("launching '%s'", cmdline);
        g_spawn_command_line_sync (cmdline,
                                   NULL,
                                   &helper_stderr,
                                   &status,
                                   NULL);
        g_free (cmdline);

        return_code = WEXITSTATUS (status);
        g_debug ("helper return code was %d", return_code);

        if (return_code == 1) {
                char **tokens;
                PolKitResult polkit_result;
                char *polkit_result_textual;
                char *polkit_action;
                guint32 xid;

                tokens = g_strsplit (helper_stderr, " ", 2);
                if (g_strv_length (tokens) != 2) {
                        g_warning ("helper return string '%s' malformed", helper_stderr);
                        g_strfreev (tokens);
                        goto out;
                }

                polkit_result_textual = tokens[0];
                polkit_action = tokens[1];

                if (!polkit_result_from_string_representation (polkit_result_textual, &polkit_result)) {
                        g_warning ("polkit_result_textual '%s' malformed", polkit_result_textual);
                        g_strfreev (tokens);
                        goto out;
                }

                g_debug ("helper refused; returned polkit_result='%s' (%d) and polkit_action='%s'", 
                         polkit_result_textual, polkit_result, polkit_action);

                xid = gdk_x11_drawable_get_xid (GDK_DRAWABLE (toplevel_window->window));

                /* now use PolicyKit-gnome to bring up an auth dialog */
                if (!dbus_g_proxy_call_with_timeout (polkit_gnome_proxy,
                                                     "ShowDialog",
                                                     INT_MAX,
                                                     &error,
                                                     /* parameters: */
                                                     G_TYPE_STRING, polkit_action,  /* action_id */
                                                     G_TYPE_UINT, xid, /* X11 window ID */
                                                     G_TYPE_INVALID,
                                                     /* return values: */
                                                     G_TYPE_BOOLEAN, &gained_privilege,
                                                     G_TYPE_INVALID)) {
                        g_warning ("Caught exception '%s'", error->message);
                        g_error_free (error);
                        g_strfreev (tokens);
                        goto out;
                }

                g_debug ("gained privilege = %d", gained_privilege);

                g_strfreev (tokens);
        }

out:
        g_free (helper_stderr);
}


static void
update_one_button (GtkWidget *button, const char *pk_action_id)
{
        PolKitCaller *pk_caller;
        PolKitAction *pk_action;
        PolKitResult pk_result;
        DBusError dbus_error;
        char *s;

        pk_caller = NULL;
        pk_action = NULL;

        pk_action = polkit_action_new ();
        polkit_action_set_action_id (pk_action, pk_action_id);

        dbus_error_init (&dbus_error);
        pk_caller = polkit_caller_new_from_pid (system_bus, getpid (), &dbus_error);
        if (pk_caller == NULL) {
                fprintf (stderr, "cannot get caller from dbus name\n");
                goto out;
        }

        pk_result = polkit_context_can_caller_do_action (pk_context, pk_action, pk_caller);

        g_debug ("for '%s' we got '%s'", pk_action_id, polkit_result_to_string_representation (pk_result));

        switch (pk_result) {
        default:
        case POLKIT_RESULT_UNKNOWN:
        case POLKIT_RESULT_NO:
                gtk_button_set_image (GTK_BUTTON (button),
                                      gtk_image_new_from_stock (GTK_STOCK_STOP, GTK_ICON_SIZE_SMALL_TOOLBAR));
                gtk_widget_set_tooltip_markup (button, _("This action is not permitted"));
                break;

        case POLKIT_RESULT_ONLY_VIA_ADMIN_AUTH:
        case POLKIT_RESULT_ONLY_VIA_ADMIN_AUTH_KEEP_SESSION:
        case POLKIT_RESULT_ONLY_VIA_ADMIN_AUTH_KEEP_ALWAYS:
        case POLKIT_RESULT_ONLY_VIA_SELF_AUTH:
        case POLKIT_RESULT_ONLY_VIA_SELF_AUTH_KEEP_SESSION:
        case POLKIT_RESULT_ONLY_VIA_SELF_AUTH_KEEP_ALWAYS:
                gtk_button_set_image (GTK_BUTTON (button),
                                      gtk_image_new_from_stock (GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_SMALL_TOOLBAR));

                s = g_strdup_printf (_("This action requires authentication (%s)"), 
                                     polkit_result_to_string_representation (pk_result));
                gtk_widget_set_tooltip_markup (button, s);
                g_free (s);
                break;

        case POLKIT_RESULT_YES:
                gtk_button_set_image (GTK_BUTTON (button),
                                      gtk_image_new_from_stock (GTK_STOCK_OK, GTK_ICON_SIZE_SMALL_TOOLBAR));
                gtk_widget_set_tooltip_markup (button, _("This action is permitted"));
                break;
        }

out:
        if (pk_action != NULL)
                polkit_action_unref (pk_action);
        if (pk_caller != NULL)
                polkit_caller_unref (pk_caller);
}

static void
update_buttons (void)
{
        update_one_button (button_frobnicate, "org.gnome.policykit.examples.frobnicate");
        update_one_button (button_tweak, "org.gnome.policykit.examples.tweak");
        update_one_button (button_twiddle, "org.gnome.policykit.examples.twiddle");
        update_one_button (button_punch, "org.gnome.policykit.examples.punch");
}

static gboolean
io_watch_have_data (GIOChannel *channel, GIOCondition condition, gpointer user_data)
{
        int fd;
        PolKitContext *pk_context = user_data;
        fd = g_io_channel_unix_get_fd (channel);
        polkit_context_io_func (pk_context, fd);
        return TRUE;
}

static int 
io_add_watch (PolKitContext *pk_context, int fd)
{
        guint id = 0;
        GIOChannel *channel;
        channel = g_io_channel_unix_new (fd);
        if (channel == NULL)
                goto out;
        id = g_io_add_watch (channel, G_IO_IN, io_watch_have_data, pk_context);
        if (id == 0) {
                g_io_channel_unref (channel);
                goto out;
        }
        g_io_channel_unref (channel);
out:
        return id;
}

static void 
io_remove_watch (PolKitContext *pk_context, int watch_id)
{
        g_source_remove (watch_id);
}


static void
pk_config_changed (PolKitContext *pk_context, void *user_data)
{
        g_debug ("PolicyKit config changed");
        update_buttons ();
}

int
main (int argc, char **argv)
{
        GMainLoop       *loop;
        DBusGConnection *session_bus;
        int              ret;
        DBusError        dbus_error;

        ret = 1;

        if (!g_thread_supported ()) {
                g_thread_init (NULL);
        }
        dbus_g_thread_init ();
        g_type_init ();
        gtk_init (&argc, &argv);

        loop = g_main_loop_new (NULL, FALSE);


        session_bus = get_session_bus ();
        if (session_bus == NULL) {
                goto out;
        }

        dbus_error_init (&dbus_error);
        system_bus = dbus_bus_get (DBUS_BUS_SYSTEM, &dbus_error);
        if (system_bus == NULL) {
                goto out;
        }

        pk_context = polkit_context_new ();
        polkit_context_set_io_watch_functions (pk_context, io_add_watch, io_remove_watch);
        polkit_context_set_config_changed (pk_context, pk_config_changed, NULL);
        if (!polkit_context_init (pk_context, NULL)) {
                goto out;
        }
        

	polkit_gnome_proxy = dbus_g_proxy_new_for_name (session_bus,
                                                        "org.gnome.PolicyKit",           /* bus name */
                                                        "/org/gnome/PolicyKit/Manager",  /* object */
                                                        "org.gnome.PolicyKit.Manager");  /* interface */

        GtkWidget *vbox;
        GtkWidget *label;
        GtkWidget *separator;
        


        toplevel_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);


        vbox = gtk_vbox_new (FALSE, 10);
        gtk_container_add (GTK_CONTAINER (toplevel_window), vbox);

        /*----------------------------------------------------------------*/

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Frobnicate</b> requires the user to authenticate."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_frobnicate = gtk_button_new_with_label (_("Frobnicate!"));
        g_signal_connect (button_frobnicate, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_frobnicate, FALSE, FALSE, 0);


        /*----------------------------------------------------------------*/
        separator = gtk_hseparator_new ();
        gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, FALSE, 0);
        /*----------------------------------------------------------------*/

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Tweak</b> requires a system administrator to authenticate."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_tweak = gtk_button_new_with_label (_("Tweak!"));
        g_signal_connect (button_tweak, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_tweak, FALSE, FALSE, 0);

        /*----------------------------------------------------------------*/
        separator = gtk_hseparator_new ();
        gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, FALSE, 0);
        /*----------------------------------------------------------------*/

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Twiddle</b> requires a system administrator to authenticate. Once authenticated, this privilege can be retained indefinitely."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_twiddle = gtk_button_new_with_label (_("Twiddle!"));
        g_signal_connect (button_twiddle, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_twiddle, FALSE, FALSE, 0);

        /*----------------------------------------------------------------*/
        separator = gtk_hseparator_new ();
        gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, FALSE, 0);
        /*----------------------------------------------------------------*/

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Punch</b> requires the user to authenticate. Once authenticated, this privilege can be retained for the remainder of the desktop session."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_punch = gtk_button_new_with_label (_("Punch!"));
        g_signal_connect (button_punch, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_punch, FALSE, FALSE, 0);

        /*----------------------------------------------------------------*/
        separator = gtk_hseparator_new ();
        gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, FALSE, 0);
        /*----------------------------------------------------------------*/

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Kick</b> can be done in different ways (this demonstrates <i>annotations</i>). Select what to kick below and then press the Kick button.."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_kick = gtk_button_new_with_label (_("Kick!"));
        g_signal_connect (button_kick, "clicked", G_CALLBACK (simple_button_cb), NULL);
        kick_combo_box = gtk_combo_box_new_text();
        gtk_combo_box_append_text (GTK_COMBO_BOX (kick_combo_box), "foo");
        gtk_combo_box_append_text (GTK_COMBO_BOX (kick_combo_box), "bar");
        gtk_combo_box_append_text (GTK_COMBO_BOX (kick_combo_box), "baz");
        gtk_combo_box_set_active (GTK_COMBO_BOX (kick_combo_box), 0);
        gtk_box_pack_start (GTK_BOX (vbox), kick_combo_box, FALSE, FALSE, 0);
        gtk_box_pack_start (GTK_BOX (vbox), button_kick, FALSE, FALSE, 0);

        /*----------------------------------------------------------------*/

        update_buttons ();

        gtk_window_set_title (GTK_WINDOW (toplevel_window), _("PolicyKit-gnome demo"));
        g_signal_connect (toplevel_window, "delete-event", exit, NULL);
        gtk_widget_show_all (toplevel_window);

        g_main_loop_run (loop);

        g_main_loop_unref (loop);
        ret = 0;
out:
        return ret;
}
